import os
import json
import hashlib
import time
import logging
try:
    from urllib.parse import urljoin
except ImportError:
    from urlparse import urljoin

from . import util
from .archive import Archive
from .session import Session, GetRequest, PutRequest
from .cache import Cache
from .package_stub import PackageStub


class Index(object):
    def __init__(self, app_name, app_version, data_path, repository_url, **kwargs):
        self.logger = logging.getLogger(__name__)

        self.app_name = app_name
        self.app_version = app_version

        if not repository_url:
            raise Exception('invalid repository_url: %s' % repository_url)

        self.data_path = data_path
        self.repository_url = repository_url

        super(Index, self).__init__(**kwargs)

    def upload(self, path):
        from boto.s3.connection import S3Connection

        access_key_id = os.environ['AWS_ACCESS_KEY_ID']
        secret_access_key = os.environ['AWS_SECRET_ACCESS_KEY']
        os.environ['S3_USE_SIGV4'] = 'True'

        # get aws/s3 upload information
        session = Session(self.app_name, self.app_version, self.data_path)
        url = urljoin(self.repository_url, '/upload')
        request = GetRequest(url)
        response = session.open(request, 'utf8')
        result = json.load(response)

        conn = S3Connection(access_key_id, secret_access_key,
                            host='s3.%s.amazonaws.com' % result['region'])
        bucket = conn.get_bucket(result['bucket'], validate=False)

        # to allow random access we upload each archive member individually
        archive = Archive(path)
        for key_name, f in archive.fileobjs().items():
            self.logger.info('preparing upload for %s', key_name)
            headers = {
                util.s3_header('md5'): hashlib.md5(f.read()).hexdigest()
            }
            f.seek(os.SEEK_SET, 0)

            self.logger.info('uploading %s...', key_name)
            key = bucket.new_key(key_name)
            key.set_contents_from_file(f, headers=headers)

        # without reindexing the index server wouldn't know that we
        # uploaded a new package
        request = PutRequest(urljoin(self.repository_url, '/reindex'))
        response = session.open(request)

        res = response.getcode() == 200
        self.logger.info('reindex %s', res)
        return res

    def update(self, max_retries=5):
        if max_retries <= 0:
            raise Exception('index server out of sync')

        request = GetRequest(urljoin(self.repository_url, '/models'))
        session = Session(self.app_name, self.app_version, self.data_path)
        cache = Cache(self.app_name, self.app_version, self.data_path)

        # remember cached packages for removal
        packages = cache.find()

        index = json.load(session.open(request, 'utf8'))

        for ident, (meta_url, etag) in index.items():
            if not cache.exists(ident, etag):
                url = urljoin(self.repository_url, meta_url)
                request = GetRequest(url)
                response = session.open(request, 'utf8')
                meta = json.load(response)

                package = PackageStub(meta['package'])
                assert ident == package.ident

                # index server's etag should match s3's etag
                if util.unquote(response.headers['etag']) != etag:
                    self.logger.info('wait for index server to sync')
                    time.sleep(3)
                    return self.update(max_retries - 1)

                assert util.unquote(response.headers['etag']) == etag
                cache.update(meta, url=url, etag=etag)

            # shrink list by one
            packages = [p for p in packages if p.ident != ident]

        # remove leftovers
        for package in packages:
            package.remove()



import os

from .package_stub import PackageStub


class DirPackage(PackageStub):

    def __init__(self, path):
        self.path = path

    def has_file(self, *path_parts):
        return os.path.isfile(os.path.join(self.path, *path_parts))

    def file_path(self, *path_parts):
        return os.path.join(self.path, *path_parts)

    def dir_path(self, *path_parts):
        return os.path.join(self.path, *path_parts)



import os
import logging
from glob import glob

from . import util
from .package_stub import PackageStub
from .archive_writer import ArchiveWriter
from .archive import Archive


class ValidationException(Exception): pass


class Recipe(PackageStub):  # package archive
    def __init__(self, recipe_path, base_path=None):
        defaults = util.json_load(os.path.join(recipe_path, "package.json"))
        super(Recipe, self).__init__(defaults)

        self.logger = logging.getLogger(__name__)

        if not recipe_path or not os.path.isdir(recipe_path):
            raise Exception('invalid recipe_path: %s' % recipe_path)

        self.recipe_path = recipe_path
        self.base_path = base_path or recipe_path

        self.include = defaults.get('include')
        self.is_valid(True)

    def is_valid(self, raise_exception=False):
        required_keys = ['name', 'version', 'include']
        for key in required_keys:
            if getattr(self, key) is None:
                if raise_exception:
                    raise ValidationException('invalid %s: %s' % (key, getattr(self, key)))
                else:
                    return False
        return True

    def build(self, archive_path):
        if os.path.isdir(archive_path):
            filename = util.archive_filename(self.name, self.version, suffix=True)
            archive_path = os.path.join(archive_path, filename)

        archive = ArchiveWriter(archive_path, base_path=self.base_path)
        self.logger.info("build %s", archive.path)

        for include in self.include:
            for path in glob(os.path.join(self.recipe_path, os.path.sep.join(include))):
                if os.path.isfile(path):
                    archive.add(path)
        archive.add_json('package', self.to_dict())
        archive.close()

        return Archive(archive_path)



import pytest

from ..cache import Cache
from ..package_stub import PackageStub
from ..package_list import (CompatiblePackageNotFoundException,
                            PackageNotFoundException)


def test_list(tmp_path):
    cache = Cache('test', '1.0.0', tmp_path)

    for i in range(1, 5):
        package = PackageStub({'name': 'abc', 'version': '%d.0.0' % i})
        cache.update({
            'archive': ['archive.gz', None],
            'package': package.to_dict()
        }, None)

    assert len(cache.find('xyz')) == 0

    assert len(cache.find()) == 4
    assert len(cache.find('')) == 4

    assert len(cache.find('abc')) == 4
    assert len(cache.find('abc ==1.0.0')) == 1
    assert len(cache.find('abc ==0.1.0')) == 0
    assert len(cache.find('abc>=1.9.0')) == 3


def test_update(tmp_path):
    cache = Cache('test', '1.0.0', tmp_path)
    package = PackageStub({'name': 'abc', 'version': '1.0.0'})

    meta = {
        'archive': ['archive.gz', None],
        'package': package.to_dict()
    }

    assert len(cache.find()) == 0

    cache.update(meta, None)

    assert len(cache.find()) == 1
    assert cache.find()[0].ident == package.ident

    with pytest.raises(CompatiblePackageNotFoundException):
        assert cache.get('abc >=1.0.1')

    assert cache.get('abc').ident == package.ident
    assert cache.get('abc ==1.0.0').ident == package.ident
    assert cache.get('abc >0.0.1').ident == package.ident

    with pytest.raises(PackageNotFoundException):
        assert cache.get('xyz')


def test_remove(tmp_path):
    cache = Cache('test', '1.0.0', tmp_path)
    package = PackageStub({'name': 'abc', 'version': '1.0.0'})

    meta = {
        'archive': ['archive.gz', None],
        'package': package.to_dict()
    }

    assert len(cache.find()) == 0

    cache.update(meta, None)

    assert len(cache.find()) == 1
    assert cache.find()[0].ident == package.ident

    package = cache.get('abc')
    cache.remove(package)

    assert len(cache.find()) == 0

    with pytest.raises(PackageNotFoundException):
        assert cache.get('abc')


def test_update_compatible(tmp_path):
    cache = Cache('test', '1.0.0', tmp_path)
    package = PackageStub({
        'name': 'abc',
        'version': '1.0.0',
        'compatibility': {
            'test': None
        }
    })

    meta = {
        'archive': ['archive.gz', None],
        'package': package.to_dict()
    }

    assert len(cache.find()) == 0

    cache.update(meta, None)

    assert len(cache.find()) == 1

    with pytest.raises(CompatiblePackageNotFoundException):
        assert cache.get('abc >=1.0.1')

    assert cache.get('abc').ident == package.ident
    assert cache.get('abc ==1.0.0').ident == package.ident
    assert cache.get('abc >0.0.1').ident == package.ident

    with pytest.raises(PackageNotFoundException):
        assert cache.get('xyz')


def test_update_incompatible(tmp_path):
    cache = Cache('test', '1.0.0', tmp_path)
    package = PackageStub({
        'name': 'abc',
        'version': '1.0.0',
        'compatibility': {
            'test': None
        }
    })

    meta = {
        'archive': ['archive.gz', None],
        'package': package.to_dict()
    }

    assert len(cache.find()) == 0

    cache.update(meta, None)

    assert len(cache.find()) == 1

    with pytest.raises(PackageNotFoundException):
        assert cache.get('xyz')


def test_update_multiple_compatible(tmp_path):
    cache = Cache('test', '5.0.0', tmp_path)

    for i in range(1, 11):
        package = PackageStub({
            'name': 'abc',
            'version': '%d.0.0' % i,  # from 1.0.0 to 10.0.0
            'compatibility': {
                'test': None
            }
        })

        meta = {
            'archive': ['archive.gz', None],
            'package': package.to_dict()
        }

        cache.update(meta, None)

    assert len(cache.find()) == 10
    assert cache.get('abc').version == '10.0.0'

    with pytest.raises(PackageNotFoundException):
        assert cache.get('xyz')


def test_update_multiple_incompatible(tmp_path):
    cache = Cache('test', '0.0.0', tmp_path)

    for i in range(1, 11):
        package = PackageStub({
            'name': 'abc',
            'version': '%d.0.0' % i,  # from 1.0.0 to 10.0.0
            'compatibility': {
                'test': None
            }
        })

        meta = {
            'archive': ['archive.gz', None],
            'package': package.to_dict()
        }

        cache.update(meta, None)

    assert len(cache.find()) == 10

    with pytest.raises(PackageNotFoundException):
        assert cache.get('xyz')



import os
import shutil
import logging

from . import util
from .package import Package
from .package_list import PackageList


class NotEnoughSpaceException(Exception): pass
class PackageAlreadyInstalledException(Exception): pass


class Pool(PackageList):

    package_class = Package

    def __init__(self, app_name, app_version, path, **kwargs):
        super(Pool, self).__init__(app_name, app_version, path, **kwargs)
        self.logger = logging.getLogger(__name__)
        self.cleanup()

    def cleanup(self):
        for filename in os.listdir(self.path):
            if filename.endswith('.tmp'):
                self.logger.info('remove %s', filename)
                shutil.rmtree(os.path.join(self.path, filename))

    def install(self, archive):
        for pkg in self.find(archive.name):
            if archive.ident == pkg.ident:
                raise PackageAlreadyInstalledException(pkg.ident)

        if not util.is_enough_space(self.path, archive.archive.size()):
            raise NotEnoughSpaceException('requires %0.2f MB' %
                                          (archive.archive.size() / 1024 / 1024))

        # remove installed versions of same package
        for pkg in self.find(archive.name):
            self.remove(pkg)

        archive_name = util.archive_filename(archive.name, archive.version)
        path = os.path.join(self.path, archive_name)

        self.logger.info('install %s', os.path.basename(path))
        archive.archive.extract_all(path + '.tmp')
        os.rename(path + '.tmp', path)

        return path



import os

from . import default
from .pool import Pool
from .util import expand_path, json_print, default_data_path
from .package import Package
from .recipe import Recipe
from .archive import Archive
from .cache import Cache
from .index import Index


# TODO support asterisks in package_strings


def install(app_name,
            app_version,
            package_name,
            data_path=None,
            repository_url=None):

    if data_path is None:
        data_path = default_data_path(app_name)
    if repository_url is None:
        repository_url = default.repository_url

    package_name = expand_path(package_name)
    data_path = expand_path(data_path)

    pool = Pool(app_name, app_version, data_path)

    if os.path.isfile(package_name):
        archive = Archive(package_name)

    else:
        index = Index(app_name, app_version, data_path, repository_url)
        index.update()

        packages = pool.find(package_name)
        if packages:
            return packages[0]

        cache = Cache(app_name, app_version, data_path)
        archive = cache.fetch(package_name)

    path = pool.install(archive)
    return Package(path=path)


def build(package_path=None,
          archive_path=None):

    if package_path is None:
        package_path = default.build_package_path

    recipe = Recipe(expand_path(package_path))
    return recipe.build(expand_path(archive_path or package_path))


def remove(app_name,
           app_version,
           package_string,
           data_path=None):

    if data_path is None:
        data_path = default_data_path(app_name)

    pool = Pool(app_name, app_version, expand_path(data_path))
    packages = pool.find(package_string)
    for pkg in packages:
        pool.remove(pkg)


def search(app_name,
           app_version,
           search_string=None,
           data_path=None,
           repository_url=None):

    if search_string is None:
        search_string = default.search_string
    if data_path is None:
        data_path = default_data_path(app_name)
    if repository_url is None:
        repository_url = default.repository_url

    # TODO make it work without data path?
    index = Index(app_name, app_version, data_path, repository_url)
    index.update()

    cache = Cache(app_name, app_version, data_path)
    packages = cache.find(search_string)
    json_print([p.ident for p in packages])
    return packages


def find(app_name,
         app_version,
         package_string=None,
         meta=None,
         cache=None,
         data_path=None):

    if package_string is None:
        package_string = default.find_package_string
    if meta is None:
        meta = default.find_meta
    if cache is None:
        cache = default.find_cache
    if data_path is None:
        data_path = default_data_path(app_name)

    cls = cache and Cache or Pool
    obj = cls(app_name, app_version, expand_path(data_path))
    packages = obj.find(package_string)
    keys = not meta and ('name', 'version') or ()
    json_print([p.to_dict(keys) for p in packages])
    return packages


def upload(app_name,
           app_version,
           package_path,
           data_path=None,
           repository_url=None):

    if data_path is None:
        data_path = default_data_path(app_name)
    if repository_url is None:
        repository_url = default.repository_url

    # TODO make it work without data path?
    index = Index(app_name, app_version, data_path, repository_url)
    return index.upload(expand_path(package_path))


def update(app_name,
           app_version,
           data_path=None,
           repository_url=None):

    if data_path is None:
        data_path = default_data_path(app_name)
    if repository_url is None:
        repository_url = default.repository_url

    index = Index(app_name, app_version, expand_path(data_path), repository_url)
    index.update()


def package(app_name,
            app_version,
            package_string,
            data_path=None):

    if data_path is None:
        data_path = default_data_path(app_name)

    pool = Pool(app_name, app_version, expand_path(data_path))
    return pool.get(package_string)


def files(app_name,
          app_version,
          package_string,
          data_path=None):

    if data_path is None:
        data_path = default_data_path(app_name)

    if os.path.isfile(package_string):
        obj = Archive(package_string)
    else:
        pool = Pool(app_name, app_version, expand_path(data_path))
        obj = pool.get(package_string)

    res = dict([(os.path.sep.join(f['path']), {'checksum': f['checksum'], 'size': f['size']})
                for f in obj.manifest])
    json_print({obj.ident: res})
    return res


def purge(app_name,
          app_version,
          cache=None,
          pool=None,
          data_path=None):

    if cache is None:
        cache = default.purge_cache
    if pool is None:
        pool = default.purge_pool
    if data_path is None:
        data_path = default_data_path(app_name)

    data_path = expand_path(data_path)

    if cache or not cache and not pool:
        Cache(app_name, app_version, data_path).purge()

    if pool or not cache and not pool:
        Pool(app_name, app_version, data_path).purge()



import os
import csv


def add_path(mod_name, path):
    record_path = get_record_path(mod_name)
    if record_path:
        full_path = '/'.join([mod_name, path])
        if not record_has_path(record_path, full_path):
            record_add_path(record_path, full_path)


def get_mod_path(mod_name):
    try:
        mod = __import__(mod_name, globals(), locals(), 0)
    except ImportError:
        return
    return os.path.join(os.path.dirname(mod.__file__))


def get_meta_path(mod_name):
    mod_path = get_mod_path(mod_name)
    if not mod_path:
        return
    path = os.path.join(mod_path, '..')
    for item in os.listdir(path):
        if item.startswith(mod_name) and item.endswith('dist-info'):
            return os.path.join(path, item)


def get_record_path(mod_name):
    meta_path = get_meta_path(mod_name)
    if not meta_path:
        return
    record_path = os.path.abspath(os.path.join(meta_path, 'RECORD'))
    if os.path.exists(record_path):
        return record_path


def record_has_path(record_path, path):
    with open(record_path) as f:
        for row in csv.reader(f):
            if row[0] == path:
                return True
    return False


def record_add_path(record_path, path):
    with open(record_path, 'a') as f:
        csvfile = csv.writer(f)
        csvfile.writerow([path, None, None])



import os

import pytest

from .. import install, build, remove, search, find, upload, update, files, purge


repository_url = os.environ.get('REPOSITORY_URL', 'https://index-staging.spacy.io')


@pytest.mark.remote
def test_upload_package(sample_package_path, tmp_path):
    archive = build(sample_package_path)
    assert os.path.isfile(archive.path)

    res = upload(None, None, archive.path, data_path=tmp_path, repository_url=repository_url)
    assert res


@pytest.mark.remote
def test_upload_package2(sample_package_path2, tmp_path):
    archive = build(sample_package_path2)
    assert os.path.isfile(archive.path)

    res = upload(None, None, archive.path, data_path=tmp_path, repository_url=repository_url)
    assert res


@pytest.mark.remote
def test_install_package(tmp_path):
    packages = find(None, None, data_path=tmp_path)
    assert len(packages) == 0

    package = install(None, None, 'test', data_path=tmp_path, repository_url=repository_url)
    assert os.path.isdir(package.path)

    packages = find(None, None, data_path=tmp_path)
    assert len(packages) == 1
    assert packages[0].ident == package.ident


@pytest.mark.remote
def test_upgrade_package(tmp_path):
    packages = find(None, None, data_path=tmp_path)
    assert len(packages) == 0

    package1 = install(None, None, 'test ==1.0.0', data_path=tmp_path, repository_url=repository_url)
    assert os.path.isdir(package1.path)

    packages = find(None, None, data_path=tmp_path)
    assert len(packages) == 1
    assert packages[0].ident == package1.ident

    package2 = install(None, None, 'test ==2.0.0', data_path=tmp_path, repository_url=repository_url)
    assert os.path.isdir(package2.path)

    packages = find(None, None, data_path=tmp_path)
    assert len(packages) == 1
    assert packages[0].ident == package2.ident


@pytest.mark.remote
def test_search_packages(tmp_path):
    assert search(None, None, data_path=tmp_path, repository_url=repository_url)
